
% Errechnet die Cholesky-Zerlegung für eine symmetrische und positiv semidefinite Matrix A.
%
% Im semidefiniten Fall wird der Parameter eps benutzt, um durch Rundungs-
% fehler entstandene Werte alpha > 0 auf der Hauptdiagonalen als 0
% anzusehen, falls alpha < eps ist.
%
% Parameter:
%   A  : symmetrische, positiv semidefinite Matrix
%   eps: Grenze für Rundungsfehler
% Rückgabe:
%   G  : linke untere Dreiecksmatrix, das "Cholesky-Dreieck", für das gilt:
%        G * G' = A
%
function [G,time,maxerr] = cholesky(A, eps)
    % nur ein Argument beim Aufruf gesetzt, setze default-Wert für eps
    if (nargin == 1)
        eps = 1e-14;
    end
    
    % Überprüfung, ob A quadratisch ist
    [n,m] = size(A);
    if (n ~= m)
        error('A is not square');
    end
    % Überprüfung auf Symmetrie
    if (~all(all(A == A')))
        error('A is not symmetric');
    end
    
    % Hauptalgorithmus -- Methode der äußeren-Produkt-Updates
    maxerr = 0;
    tic;
    for k = 1:n
        alpha = A(k,k);
        
        % Wert auf der Hauptdiagonalen gibt Aufschluss über Definitheit
        if (alpha < 0)
            error('A is not positive definite: G(%d,%d) = %g < 0', ...
                  k, k, alpha);
            return
        end
        
        % alpha < eps wird als 0 angesehen
        if (alpha < eps)
            % In diesem Fall ist sind die zugehörige Zeile und Spalte 0 zu
            % setzen. Speichere zusätzlich den maximal tolerierten Fehler.
            maxerr = alpha;
            A(k:n,k) = 0;
            A(k,k:n) = 0;
        else
            beta = sqrt(alpha);
            A(k,k) = beta;
            A(k+1:n,k) = A(k+1:n,k) / beta;
            % äußeres-Produkt-Update
            for j = k+1:n
                A(j:n,j) = A(j:n,j) - A(j:n,k) * A(j,k);
            end
        end
    end
    time = toc;
    
    % Erstelle Rückgabewert: linke untere Dreicksmatrix aus den Werten in A
    G = zeros(n,n);
    for k = 1:n
        G(k:n,k) = A(k:n,k);
    end
end

% pascal(k)                  : Pascalsches Dreieck in Matrix
%   k = 1:30, dann Werte zu groß -> Rundungsfehler
% full(gallery('poisson', k)): dünnbesetzte sparse-Matrix, block-tridiagonal
% gallery('moler', k)
% gallery('gcdmat', k)       : A(i,j) = gcd(i,j) pos. def.; A.^r pos. semidef

